

#define FIXFONT

#include "common.h"
#include "cs.h"
#include "lcd.h"



//Control-Port
#define LCD_C_PORT	PORTD
#define LCD_C_DDR	DDRD
#define LCD_C_RD	2
#define LCD_C_WR	3
#define LCD_C_RS	4
#define LCD_C_CS	5
#define LCD_C_RES	6
//Data-Port
#define LCD_D_PORT	PORTA
#define LCD_D_DDR	DDRA
#define LCD_D_PIN	PINA



#define LCD_R	0
#define LCD_D	1

#define LCDCR1_ADC		1
#define LCDCR1_HOLT		2
#define LCDCR1_REV		4
#define LCDCR1_AMP		8
#define LCDCR1_PWR		16
#define LCDCR1_STBY		32
#define LCDCR1_DISP		64
#define LCDCR1_RMW		128

#define LCDCR16_CM0		32
#define LCDCR16_CM1		64

#define LCDCR17_MON		4


unsigned char str_x, str_y, px, py, pen = 1, cr, ff;
unsigned char fb, col, prp = 1, chh = 1;
unsigned char lcd_txtbox_w, lcd_txtbox_h;

void lcd_xy(unsigned char x, unsigned char y);
void lcd_setp(uint16_t x, uint16_t y);


void _write(unsigned char adr, unsigned char data)
{
	uint8_t sr = SREG;
	cli();

	LCD_C_PORT &= ~_BV(LCD_C_CS);

	LCD_C_PORT &= ~_BV(LCD_C_WR);
	if(adr & 0x01)
		LCD_C_PORT |= _BV(LCD_C_RS);

	LCD_D_DDR = 0xFF;

	LCD_D_PORT = data;
	nop();
	nop();

	LCD_C_PORT |= _BV(LCD_C_WR);
	LCD_C_PORT &= ~_BV(LCD_C_RS);
	LCD_C_PORT |= _BV(LCD_C_CS);
	LCD_D_DDR = 0;

	SREG = sr;
}

unsigned char _read(unsigned char adr)
{
	unsigned char data;

	uint8_t sr = SREG;
	cli();

	LCD_D_PORT = 0xFF;

	LCD_C_PORT &= ~_BV(LCD_C_CS);

	LCD_C_PORT &= ~_BV(LCD_C_RD);
	if(adr & 0x01)
		LCD_C_PORT |= _BV(LCD_C_RS);

	nop();
	nop();
	data = LCD_D_PIN;

	LCD_C_PORT &= ~_BV(LCD_C_RS);
	LCD_C_PORT |= _BV(LCD_C_RD);
	LCD_C_PORT |= _BV(LCD_C_CS);

	SREG = sr;

	return data;
}

void lcd_reg(unsigned char reg)
{
	_write(LCD_R, reg);
}

void lcd_data(unsigned char data)
{
	_write(LCD_D, data);
}


void lcd_wr(unsigned char reg, unsigned char data)
{
	_write(LCD_R, reg);
	_write(LCD_D, data);
}

unsigned char lcd_rd(void)
{
	_write(LCD_R, 4);
	_read(LCD_D);
	return _read(LCD_D);
}

void lcd_init()
{
	LCD_D_DDR = 0;

	LCD_C_DDR |= _BV(LCD_C_RES);
	LCD_C_PORT &= ~_BV(LCD_C_RES);

	LCD_C_PORT |= _BV(LCD_C_RD);
	LCD_C_PORT |= _BV(LCD_C_WR);
	LCD_C_PORT &= ~_BV(LCD_C_RS);
	LCD_C_PORT |= _BV(LCD_C_CS);

	LCD_C_DDR |= _BV(LCD_C_RD);
	LCD_C_DDR |= _BV(LCD_C_WR);
	LCD_C_DDR |= _BV(LCD_C_RS);
	LCD_C_DDR |= _BV(LCD_C_CS);

	LCD_C_PORT |= _BV(LCD_C_RES);

	lcd_wr(0, LCDCR1_AMP | LCDCR1_PWR | LCDCR1_DISP | LCDCR1_ADC | LCDCR1_RMW);
	lcd_wr(1, 0);
	//lcd_contrast(setup_get_lcd_contr());
	lcd_wr(17, LCDCR17_MON);

	lcd_clr();
}

void lcd_contrast(uint8_t contr)
{
	lcd_wr(16, 0x00 | (contr & 0x1F));
}

void lcd_clr(void)
{
	unsigned char x;
	int i;
	for(x = 0; x < 20; x++)
	{
		lcd_xy(x, 0);
		lcd_reg(4);
		for(i = 0; i < 100; i++)
			lcd_data(0x00);
	}
}

void lcd_xy(unsigned char x, unsigned char y)
{
	lcd_wr(2, x << 1);
	lcd_wr(3, y);
}

void lcd_pos(unsigned char x, unsigned char y)
{
	px = str_x = x;
	py = str_y = y;
}

void lcd_set_bold(unsigned char b)
{
	fb = b;
}

void lcd_lineto(uint8_t x, uint8_t y)
{
	uint16_t tmp, x2 = x, y2 = y;
	uint8_t b = 0, bx = 0, by = 0; 
	if(x < px)
	{
		tmp = x;
		x = px;
		px = tmp;
	}
	if(y < py)
	{
		tmp = y;
		y = py;
		py = tmp;
	}

	uint16_t dx = x - px;
	uint16_t dy = y - py;
	uint8_t* p1_1 = &py;
	uint8_t* p1_2 = &y;
	uint8_t* p2_1 = &px;
	int16_t e = -dy;
	int16_t d = 2 * dx;
	if(dy < dx)
	{
		e = -dx;
		d = 2 * dy;
		p1_1 = &px;
		p1_2 = &x;
		p2_1 = &py;
	}
	int16_t s = 2 * e;
	for(;;)
	{
		if((px / 8 != bx) || (py != by) || (*p1_1 == *p1_2))
		{
			lcd_xy(bx, by);
			if(pen)
				b |= lcd_rd();
			else
			{
				b = ~b;
				b &= lcd_rd();
			}
			lcd_reg(4);
			lcd_data(b);
			bx = px / 8;
			by = py;
			b = (0x80 >> (px % 8));
		}
		else
			b |= (0x80 >> (px % 8));
		if(*p1_1 == *p1_2)
			break;
		(*p1_1)++;
		e += d;
		if(e > 0)
		{
			(*p2_1)++;
			e += s;
		}
	}
	lcd_setp(px, py);
	px = x2;
	py = y2;
}

void lcd_rect(unsigned char w, unsigned char h)
{
	unsigned char x0 = px;
	unsigned char y0 = py;
	lcd_lineto(x0 + w, y0);
	lcd_lineto(x0 + w, y0 + h);
	lcd_lineto(x0, y0 + h);
	lcd_lineto(x0, y0);
}

void lcd_frect(uint8_t w, uint8_t h, uint8_t p)
{
	uint8_t i, x = px, y = py;
	pen = p;
	for(i = y; i < (y + h); i++)
	{
		lcd_pos(x, i);
		lcd_lineto(x + w, i);
	}
	pen = 1;
	lcd_pos(x, y);
}


void lcd_txtbox(unsigned char xmax, unsigned char ymax)
{
	lcd_txtbox_w = xmax;
	lcd_txtbox_h = ymax;
}

void lcd_pen(uint8_t p)
{
	pen = p;
}

void lcd_prp(uint8_t p)
{
	prp = p;
}

void lcd_chh(uint8_t h)
{
	chh = h;
}

uchar lcd_strw(char* s)
{
	uchar w = 0;
	while(*s != 0)
	{
		w += 6;
		s++;
	}
	return w;
}

uchar lcd_charh()
{
	return 13;
}


#ifdef FIXFONT
uchar wr_char(unsigned char c, uint wmax, uint hmax, uchar mode)
{
	uint8_t d, b, i, j, k, l, x, _x, xs = 0;
	uint16_t idx = (unsigned int)font + ((c - 0x20) * 6);
	uint8_t _fb = fb + 1;
	uint8_t w = prp ? 0 : 6;
	uint8_t fr[8];

	if(!wmax)
		wmax = 160;

	for(i = 0; i < 6; i++)
	{
		d = __LPM(idx + i);
		for(j = 0; j < 8; j++)
		{
			fr[j] >>= 1;
			fr[j] |= d & 0x80;
			if(d & 0x80)
				xs = 5 - i;
			if((w == 0) && (d & 0x80))
				w = 6 - i;
			d <<= 1;
		}
	}

	if(prp)
	{
		w -= xs;
		if(c == ' ')
			w = 2;
		for(j = 0; j < 8; j++)
			fr[j] <<= xs;
	}


	for(i = 0; i < 8; i++)
	{
		if(wmax && (str_x + w * _fb) >= wmax)
			return 0;
		_x = str_x / 8;
		lcd_xy(_x, str_y + (i * (_fb * chh)));
		b = lcd_rd();
		x = str_x % 8;
		for(j = 0; j < w; j++)
		{
			for(l = 0; l < _fb; l++)
			{
				b &= ~(0x80 >> (x));
				b |= (fr[i] & 0x80) >> (x);
				x++;
				if((x == 8) || ((j == w - 1) && (l == _fb - 1)))
				{
					for(k = 0; k < (_fb * chh); k++)
						lcd_data(b);
					_x++;
					lcd_xy(_x, str_y + (i * (_fb * chh)));
					b = lcd_rd();
					x = 0;
				}
			}
			fr[i] <<= 1;
		}
	}
	str_x += w * _fb + 1;

	return 1;

}
#else
uchar wr_char(unsigned char c, uint wmax, uint hmax, uchar mode)
{
	unsigned int idx;
	unsigned char i;
	unsigned char z = 0;
	unsigned char clen;
	unsigned char ch;
	uint x = str_x;

	idx = pgm_read_word_near(chsidx + (c - 0x20));
	idx += (unsigned int)chs;
	clen = pgm_read_byte_near(idx);
	if(c == 0x20)
	{
		str_x += 2;
		return 1;
	}
	if(c == '\n')
	{
		str_x = px;
		py += lcd_charh();
		if(mode == 3)
			py += 5;
		if(!cr)
			return 0;
		return 1;
	}
	else if(mode == 4)
		return 1;

	if(c == '\t')
	{
		str_x += TAB_WIDTH - (str_x % TAB_WIDTH);
		return 0;
	}
	
	idx++;

	if(ff)
		str_x = str_x + (ff / 2) - (clen / 2);

	if(wmax && ((str_x + clen) >= wmax))
	{
		if(mode == 3)
			return 1;
		py += lcd_charh();
		str_x = px;
		if(!cr)
		{
			return 0;
		}
	}

	
	if((mode && (mode != 3)) || (hmax && ((py + lcd_charh()) >= hmax)))
	{
		str_x += clen;
		clen = 0;
		if(!cr && (mode != 2))
			return 0;
		if(!mode)
			mode = 99;
	}
	
	for(i = clen; i > 0; i--)
	{
		ch = pgm_read_byte_near(idx + (i - 1));

		for(z = 0; z < 13; z++)
		{
			if((ch & 0x80) != 0)
				lcd_setp(str_x, py + z);
			ch <<= 1;
			if(z == 7)
				ch = pgm_read_byte_near(idx + clen + (i - 1));
		}
		str_x++;
	}

	if(ff)
		str_x = x + ff;

	str_x++;
	if(mode == 99)
		return 0;
	return 1;
}
#endif

uchar lcd_char(unsigned char c)
{
	return wr_char(c, lcd_txtbox_w, lcd_txtbox_h, 0);
}

char* lcd_str(char* s)
{
	while((*s != 0) && lcd_char(*s))
		s++;
	return s;
}

void lcd_pstr(PGM_P s)
{
	unsigned char c;

	while((c = pgm_read_byte_near(s)) != 0)
	{
		lcd_char(c);
		s++;
	}
}

void lcd_setp(uint16_t x, uint16_t y)
{
	uint8_t b;
	lcd_xy(x / 8, y);
	b = lcd_rd();
	if(pen)
		b |= 0x80 >> (x % 8);
	else
		b &= ~(0x80 >> (x % 8));
	lcd_data(b);
}


